import numpy as np

def gaussian(x, width, amplitude, position):
    '''
    Calculates the value of a Gaussian peak at the given x-values.
    
    Args:
        x (numpy.ndarray): x-values at which to evaluate the peak.
        width (float): Width of the peak.
        amplitude (float): Amplitude of the peak.
        position (float): Position of the peak.
    '''
    return amplitude * np.exp(-0.5 * ((x - position) / width)**2)

def lorentzian(x, width, amplitude, position):
    '''
    Calculates the value of a Lorentzian peak at the given x-values.
    
    Args:
        x (numpy.ndarray): x-values at which to evaluate the peak.
        width (float): Width of the peak.
        amplitude (float): Amplitude of the peak.
        position (float): Position of the peak.
    '''
    return amplitude / (1 + ((x - position) / width)**2)

class Peak:
    '''
    A class representing a peak in a spectrum.
    
    Notably, the user initializes a peak, and then uses a `Deconvolution` object to fit the peak to a spectrum.
    The peak can be plotted using the `plot` method.
    '''
    
    def __init__(self, peak_type, width, amplitude, position, constrain_width=(0,np.inf), constrain_amplitude=(0,np.inf), constrain_position=(-np.inf,np.inf)):
        '''
        Constructor for the Peak class.
        
        Args:
            peak_type (str): Type of peak to be used. Of 'gaussian' or 'lorentzian'.
            width (float): Initial width of the peak.
            amplitude (float): Initial amplitude of the peak.
            position (float): Initial position of the peak.
            constrain_width (tuple): Tuple of the form (min, max); range for the width of the peak.
            constrain_amplitude (tuple): Tuple of the form (min, max); range for the amplitude of the peak.
            constrain_position (tuple): Tuple of the form (min, max); range for the position of the peak.
        '''
        
        assert peak_type in ['gaussian', 'lorentzian'], "peak_type must be 'gaussian' or 'lorentzian'."
        
        self.peak_type = peak_type
        self.width = width
        self.amplitude = amplitude
        self.position = position
        self.constrain_width = constrain_width
        self.constrain_amplitude = constrain_amplitude
        self.constrain_position = constrain_position
    
    def __call__(self, x):
        '''
        Calculates the value of the peak at the given x-values.
        
        Args:
            x (numpy.ndarray): x-values at which to evaluate the peak.
        '''
        if self.peak_type == 'gaussian':
            return gaussian(x, self.width, self.amplitude, self.position)
        elif self.peak_type == 'lorentzian':
            return lorentzian(x, self.width, self.amplitude, self.position)
        else:
            raise ValueError("`peak_type` must be 'gaussian' or 'lorentzian'.")
    
    def set_perr(self, width_perr, amplitude_perr, position_perr):
        self.width_perr = width_perr
        self.amplitude_perr = amplitude_perr
        self.position_perr = position_perr
    
    def get_fit_error(self, x):
        '''
        Calculates the error bars of the peak at the given x-values given 1-sigma from the fit, as set by `set_perr`.
        
        Returns a pair of the upper (`ub`) and lower (`lb`) bounds of the peak where `ub` and `lb` are the same shape as `x`.
        
        Args:
            x (numpy.ndarray): x-values at which to evaluate the peak.
        '''
        if self.peak_type == 'gaussian':
            ub = gaussian(x, self.width + self.width_perr, self.amplitude + self.amplitude_perr, self.position + self.position_perr)
            lb = gaussian(x, self.width - self.width_perr, self.amplitude - self.amplitude_perr, self.position - self.position_perr)
            return ub, lb
        elif self.peak_type == 'lorentzian':
            ub = lorentzian(x, self.width + self.width_perr, self.amplitude + self.amplitude_perr, self.position + self.position_perr)
            lb = lorentzian(x, self.width - self.width_perr, self.amplitude - self.amplitude_perr, self.position - self.position_perr)
            return ub, lb
        else:
            raise ValueError("`peak_type` must be 'gaussian' or 'lorentzian'.")

    def plot(self, ax, x=None, baseline=None, show_error=False, **kwargs):
        '''
        Plots the peak on the given axis.
        
        Args:
            ax (matplotlib.axes.Axes): Axis on which to plot the peak.
            x (numpy.ndarray): x-values at which to evaluate the peak. If None, x-values are generated near the position.
            baseline (Baseline): Baseline object to be plotted along with the peak. If None, no baseline is plotted.
            show_error (bool): Whether or not to plot error bars on the peak.
            **kwargs: Additional matplotlib keyword arguments to be passed to the plot function.
        '''
        if x is None:
            fwhm = self.get_fwhm()
            x = np.linspace(self.position - 4*fwhm, self.position + 4*fwhm, 1000)
        
        if show_error:
            ub, lb = self.get_fit_error(x)
            if baseline is not None:
                ub += baseline(x)
                lb += baseline(x)
            ax.fill_between(x, lb, ub, alpha=0.5, **kwargs)
        
        y = self(x)
        if baseline is not None:
            y += baseline(x)
        ax.plot(x, y, **kwargs)
        
    def get_fwhm(self):
        '''
        Gets the full width at half maximum of the peak.
        '''
        if self.peak_type == 'gaussian':
            # 2 sqrt(2 ln(2)) ~= 2.35482005
            return 2.35482005 * self.width
        elif self.peak_type == 'lorentzian':
            return 2 * self.width
        else:
            raise ValueError("peak_type must be 'gaussian' or 'lorentzian'.")



import csv
import requests
import numpy as np
import jcamp

def import_ascii_file(fname, delim=None):
    '''
    Import x (abscissa), y (ordinate) spectrum data from file. Assumes two column ascii format where the first column is x and the second column is y.
    
    Args:
        fname (str): Path to file.
        delim (str): Delimiter used in file. If None, will attempt to guess.
    '''
    try:
        f = open(fname, 'r')
    except IOError:
        print(f'Unable to open file at {fname}')
        return None, None
    except Exception as e:
        print(f'Error: {e}')
        return None, None
    else:
        with f:
            fc = f.read()
            if delim is None:
                sniffer = csv.Sniffer()
                delimiter = sniffer.sniff(fc).delimiter
                x, y = np.transpose(np.genfromtxt(fc.splitlines(), delimiter=delimiter))
            else:
                x, y = np.transpose(np.genfromtxt(fc.splitlines(), delimiter=delim))
            
            if x is None or y is None:
                print(f'Error: Unable to parse data from {fname}')
                return None, None

            # sort y w.r.t. x
            x_ind = np.argsort(x)
            x, y = x[x_ind], y[x_ind]
            return x, y

def import_jcamp_file(fname):
    '''
    Import spectrum data from JCAMP file.
    
    !!! note
        JCAMP-DX files are loaded using the [`jcamp`](https://pypi.org/project/jcamp/) package.
        Unlike ASCII files which are assumed to have a single spectrum, JCAMP files can contain multiple spectra, and should be handled accordingly by the user.
    
    Args:
        fname (str): Path to file.
    '''
    return jcamp.jcamp_readfile(fname)

def import_ascii_url(url, delim=None):
    '''
    Import x (abscissa), y (ordinate) spectrum data from URL. Assumes two column ascii format where the first column is x and the second column is y.
    
    Args:
        url (str): URL to file.
        delim (str): Delimiter used in file. If None, will attempt to guess.
    '''
    try:
        response = requests.get(url)
        if delim is None:
            sniffer = csv.Sniffer()
            delimiter = sniffer.sniff(response.text).delimiter
            x, y = np.transpose(np.genfromtxt(response.text.splitlines(), delimiter=delimiter))
        else:
            x, y = np.transpose(np.genfromtxt(response.text.splitlines(), delimiter=delim))

        if x is None or y is None:
            print(f'Error: Unable to parse data from {url}')
            return None, None

        # sort y w.r.t. x
        x_ind = np.argsort(x)
        x, y = x[x_ind], y[x_ind]
        return x, y
    except IOError:
        print(f'Unable to open file at {url}')
        return None, None
    except Exception as e:
        print(f'Error: {e}')
        return None, None

def import_jcamp_url(url):
    '''
    Import spectrum data from JCAMP file at URL.
    
    !!! note
        JCAMP-DX files are loaded using the [`jcamp`](https://pypi.org/project/jcamp/) package.
        Unlike ASCII files which are assumed to have a single spectrum, JCAMP files can contain multiple spectra, and should be handled accordingly by the user.
    
    Args:
        url (str): URL to file.
    '''
    try:
        response = requests.get(url)
        return jcamp.jcamp_readfile(response.text)
    except IOError:
        print(f'Unable to open file at {url}')
        return None
    except Exception as e: 
        print(f'Error: {e}')
        return None

def normalize_data(x, y, range=None):
    '''
    Normalize y data to range [0, 1].
    
    If range is not None, normalize y data to range [0, 1] within the specified x range.
    
    x and y must have the same shape.
    
    Args:
        x (np.array): x data.
        y (np.array): y data.
        range (tuple): x range to normalize y data within. If None, normalize y data to range [0, 1].
    '''
    if range is None:
        return x, (y - np.min(y)) / (np.max(y) - np.min(y))
    diffs1 = np.abs(x - range[0])
    diffs2 = np.abs(x - range[1])
    i1 = np.argmin(diffs1)
    i2 = np.argmin(diffs2)
    return x, (y - np.min(y[i1:i2])) / (np.max(y[i1:i2]) - np.min(y[i1:i2]))

def crop(x, y, range):
    '''
    Crop x, y data to specified x range. `range` must be a pair of values in the domain of x.
    
    x and y must have the same shape.
    
    Args:
        x (np.array): x data.
        y (np.array): y data.
        range (tuple): x range to crop x, y data to, in the form (x_min, x_max).
    '''
    diffs1 = np.abs(x - range[0])
    diffs2 = np.abs(x - range[1])
    i1 = np.argmin(diffs1)
    i2 = np.argmin(diffs2)
    x = x[i1:i2]
    y = y[i1:i2]
    return x, y

def nearest(x, y, x0):
    '''
    Find the y value in y that corresponds to the nearest x value to x0.
    
    Useful e.g. for identifying nodes for baseline correction that exist on the spectrum.
    
    Args:
        x (np.array): x data.
        y (np.array): y data.
        x0 (float): x value to find nearest y value for.
    '''
    idx = (np.abs(x-x0)).argmin()
    return y[idx]

def nearest_points(x, y, x0):
    '''
    Find the y values in y that correspond to the nearest x values to x0.
    
    Useful e.g. for identifying nodes for baseline correction that exist on the spectrum.
    
    Args:
        x (np.array): x data.
        y (np.array): y data.
        x0 (np.array): x values to find nearest y values for.
    '''
    return np.array([nearest(x, y, x0i) for x0i in x0])




import numpy as np
from scipy.interpolate import CubicSpline

class IdentityBaseline:
    '''
    No baseline correction, or the identity baseline
    '''
    def __init__(self):
        pass
    
    def __call__(self, x):
        '''
        Calculates the value of the baseline at the given x-values.
        
        Args:
            x (numpy.ndarray): x-values at which to evaluate the baseline.
        '''
        return np.zeros_like(x)

class CubicSplineBaseline:
    '''
    Baseline represented by a cubic spline.
    '''
    def __init__(self, x, y, bc_type='not-a-knot'):
        '''
        Constructor for the cubic spline baseline.
        
        Args:
            x (numpy.ndarray): x-values of the cubic spline.
            y (numpy.ndarray): y-values of the cubic spline.
            bc_type (str): Type of boundary condition to use. Default is 'not-a-knot'. See `scipy.interpolate.CubicSpline` for more information.
        '''
        self.spline = CubicSpline(x, y)
    
    def __call__(self, x):
        '''
        Calculates the value of the baseline at the given x-values.
        
        Args:
            x (numpy.ndarray): x-values at which to evaluate the baseline.
        '''
        return self.spline(x)

class PiecewiseLinearBaseline:
    '''
    Baseline represented by a piecewise linear function.
    '''
    def __init__(self, x, y):
        '''
        Constructor for the piecewise linear baseline. 
        
        Any values outside of the range of the provided set of vertices are set to the value at the nearest edge vertex.
        
        Args:
            x (numpy.ndarray): x-values of the vertices of the piecewise linear function
            y (numpy.ndarray): y-values of the vertices of the piecewise linear function
        '''
        self.x = x
        self.y = y
    
    def __call__(self, x):
        '''
        Calculates the value of the baseline at the given x-values.
        
        Args:
            x (numpy.ndarray): x-values at which to evaluate the baseline.
        '''
        return np.interp(x, self.x, self.y)
        
    
    


import numpy as np
import copy
from scipy.optimize import curve_fit
from tabulate import tabulate

from .peak import gaussian, lorentzian, Peak
from .baseline import IdentityBaseline

class Deconvolution:
    '''
    A class representing a deconvolution of a spectrum.
    '''
    
    def __init__(self, peaks, baseline=IdentityBaseline()):
        '''
        Constructor for the Deconvolution class.
        
        Args:
            peaks (list): List of Peak objects to be fit to the spectrum.
            baseline (Baseline): Baseline for fitting peaks. Default is the identity baseline.
        '''
        self.peaks = [copy.deepcopy(peak) for peak in peaks]
        self.baseline = baseline
        
        self.N = len(peaks)
    
    def fit(self, x, y):
        '''
        Fits the peaks to the spectrum.
        
        Args:
            x (numpy.ndarray): x-values at which to evaluate the spectrum.
            y (numpy.ndarray): y-values of the spectrum.
        '''
        base = self.baseline(x)
        
        # rescale so numerically feasible
        scale_x = np.max(x) - np.min(x)
        
        fit_params = np.zeros(3*self.N)
        min_bounds = np.zeros_like(fit_params)
        max_bounds = np.zeros_like(fit_params)
        
        for i, peak in enumerate(self.peaks):
            fit_params[3*i:3*i+3] = [peak.width / scale_x, peak.amplitude, peak.position / scale_x]
            
            min_bounds[3*i:3*i+3] = [peak.constrain_width[0] / scale_x, peak.constrain_amplitude[0], peak.constrain_position[0] / scale_x]
            max_bounds[3*i:3*i+3] = [peak.constrain_width[1] / scale_x, peak.constrain_amplitude[1], peak.constrain_position[1] / scale_x]

        def optim_func(x0, *params):
            '''
            Function to be minimized in the optimization process.
            
            Args:
                x (numpy.ndarray): x-values at which to evaluate the function.
                *params: Parameters to be optimized.
            '''
            y0 = np.zeros_like(x)
            for i in range(0, len(params), 3):
                peak = Peak(self.peaks[int(i // 3)].peak_type, *params[i:i+3])
                y0 += peak(x0)
            return y0

        self.fit_params, self.pcov = curve_fit(optim_func, x / scale_x, y - base, p0=fit_params, bounds=(min_bounds, max_bounds), maxfev=100000)
        self.perr = np.sqrt(np.diag(self.pcov))
        
        for i in range(self.N):
            self.peaks[i].width, self.peaks[i].amplitude, self.peaks[i].position = self.fit_params[3*i:3*i+3]
            
            fp = self.fit_params[3*i:3*i+3]
            
            self.peaks[i].width *= scale_x
            self.peaks[i].position *= scale_x
            
            perr = self.perr[3*i:3*i+3]
            
            self.peaks[i].set_perr(perr[0] * scale_x, perr[1], perr[2] * scale_x)
        
        print(self)
    
    def __str__(self):
        '''
        Returns a string representation of the deconvolution.
        '''
        return tabulate(
            [[peak.peak_type, peak.width, peak.amplitude, peak.position] for peak in self.peaks],
            headers=['Peak Type', 'Width', 'Amplitude', 'Position']
        )

    def __call__(self, x):
        '''
        Calculates the value of the deconvolution at the given x-values.
        
        Args:
            x (numpy.ndarray): x-values at which to evaluate the deconvolution.
        '''
        y = np.zeros_like(x)
        for i in range(0, self.N*3, 3):
            y += self.peaks[int(i // 3)](x)
        return y

    def plot(self, ax, x, y=None, show_base=True, show_peaks=True, show_error=False, **kwargs):
        '''
        Plots the deconvolution on the given axis.
        
        Args:
            ax (matplotlib.axes.Axes): Axis on which to plot the deconvolution.
            x (numpy.ndarray): x-values at which to evaluate the deconvolution.
            y (numpy.ndarray): y-values of the ground truth spectrum. Optional.
            show_base (bool): Whether or not to plot the baseline.
            show_peaks (bool): Whether or not to plot the individual peaks.
            show_error (bool): Whether or not to plot the error bars.
            **kwargs: Additional matplotlib keyword arguments to be passed to the plot.
        '''
        if show_peaks:
            for peak in self.peaks:
                peak.plot(ax, x, baseline=self.baseline if show_base else None, show_error=show_error, color='r', ls='--', **kwargs)
                
        y_pred = self(x)
        if show_base:
            y_pred += self.baseline(x)
        ax.plot(x, y_pred, c='b', ls='--', label='Deconvolution', **kwargs)
        
        if y is not None:
            ax.plot(x, y, label='Ground Truth', c='k')
        


__version__ = "1.0.1"

from .peak import Peak, gaussian, lorentzian
from .deconvolution import Deconvolution
from .baseline import PiecewiseLinearBaseline, IdentityBaseline, CubicSplineBaseline
from .filereader import import_ascii_file, import_ascii_url, crop, normalize_data, nearest_points

from . import baseline
from . import deconvolution
from . import filereader
from . import peak



from setuptools import setup, find_packages

setup(
    name='pypeakify',
    author='mscloudlab',
    author_email='mscloudlab@gmail.com',
    description='General purpose peak deconvolution package for Python.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    packages=find_packages(where='src'),
    install_requires=[
        'numpy',
        'scipy',
        'matplotlib',
        'tabulate',
        'requests',
        'jcamp'
    ],
    package_dir={'': 'src'},
    classifiers=[
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)
